local Me=CreateFrame("Frame","PrattleGUI")
local master=nil
local L={}
local masterL={}
local unsavedchanges=false
local visiblepanel=nil
local queuedRequest=nil
local ready=false
local mainpanel="PrattleFrame"
local db={}
local dbpC={}
local config={}
local playername=UnitName("player")
local defaultlanguage=1
local guidebug=false
local editback={0.5,0.5,0.5,1}

-- --------------------------------------------------------------------------
-- INITIALISIERUNG
-- --------------------------------------------------------------------------
function Me:initialize(prattlemaster,langarray)
	-- Verknpfung zu PrattleMaster und Language-Array zuweisen
	master=prattlemaster
	masterL=langarray
	if Lang.PrattleGUI then 
		L=Lang.PrattleGUI
	else
		master:Print("failure to load GUI language file for "..GetLocale(),nil,"ERROR")
	end
	-- Widget-Texte festlegen														-- BUTTONS
	for button,caption in pairs(L.button) do
		local widget=Me:Widget("Button_"..button)
		if widget then	-- Abort und Ignore werden hier nicht gefunden (siehe Unsaved)
			widget:SetText(caption)
			widget.tooltipText=L.tooltip[button]
		end
	end
	Me:Widget("Release"):SetText(master:getRelease())

	for i=1, GetNumLanguages() do 
		if GetLanguageByIndex(i)==GetDefaultLanguage("player") then defaultlang=i end
	end
	
	ready=true
	return Me
end

-- SET DEBUG (aufgerufen von PrattleMaster)
function Me:SetDebug(set) guidebug=set end

-- --------------------------------------------------------------------------
-- FRAME CALL
-- --------------------------------------------------------------------------
function Me:FrameCall(request,parm1,parm2)
	if guidebug then master:Debug("GUI FrameCall rq="..Me:tell(request)..",p1="..Me:tell(parm1)..",p2="..Me:tell(parm2)) end
	if not master or not ready then 													-- ignore request while initialising
	elseif request=="EDITBOX" then Me:EditboxChange(parm1,parm2)						-- EDITBOX
	elseif request=="BUTTON" then Me:ButtonClicked(parm1,parm2)							-- BUTTON
	elseif request=="SLIDE" then Me:Slide(parm1,parm2)									-- SLIDE
	elseif request=="TOOLTIP" then														-- TOOLTIP
		GameTooltip:SetOwner(parm1, "ANCHOR_RIGHT")
		if parm1.tooltipText then
			GameTooltip:SetText(parm1.tooltipText, 1.0, 0.8, 0.0, 1 )
			GameTooltip:Show();
		end
	else
		master:Print(master:gsub(L.msg.wrongrequest,"<req>",request))
	end
end

-- --------------------------------------------------------------------------
-- BUTTON CLICKED
-- --------------------------------------------------------------------------
function Me:ButtonClicked(button,parm)
	if guidebug then master:Debug("GUI ButtonClicked "..Me:tell(button)..","..Me:tell(parm)) end
	local parse={strsplit("_",button:GetName())}
	if strupper(parse[#parse])=="SAVE" then Me:FinishUnsaved("SAVE")					-- SAVE
	elseif strupper(parse[#parse])=="IGNORE" then Me:FinishUnsaved("IGNORE")			-- IGNORE
	elseif strupper(parse[#parse])=="ABORT" then Me:FinishUnsaved("ABORT")				-- ABORT
	elseif strupper(parse[#parse])=="CLOSE" then Me:Route("CLOSE")						-- CLOSE
	elseif parse[#parse-1]=="Button" and parse[#parse-2]==mainpanel then				-- Hauptmen Buttons
		Me:Route(parse[#parse])
	elseif parse[#parse]=="UP" or parse[#parse]=="DOWN" then							-- Scroll Up or down in Texts
		local scroll=parse[#parse-1]..parse[#parse]
		local top=nil
		if scroll=="FirstDOWN" and config.top1+15<=config.max1 then top={"top1",4}
		elseif scroll=="FirstUP" and config.top1>1 then top={"top1",4}
		elseif scroll=="SecondDOWN" and config.top2+7<=config.max2 then top={"top2",4}
		elseif scroll=="SecondUP" and config.top2>1 then top={"top2",4}
		elseif scroll=="TextsDOWN" and config.topt+9<=config.maxt then top={"topt",1}
		elseif scroll=="TextsUP" and config.topt>1 then top={"topt",1}
		elseif scroll=="ReplacementsDOWN" and config.topt+14<=config.maxt then top={"topt",1}
		elseif scroll=="ReplacementsUP" and config.topt>1 then top={"topt",1}
		end
		if top==nil then master:Sound("BOING")
		else
			if parse[#parse]=="UP" then config[top[1]]=config[top[1]]-top[2]
			else config[top[1]]=config[top[1]]+top[2] end
			if config.main=="Texts" then Me:TextShow()
			elseif config.main=="Replacements" then Me:ReplacementsShow()
			end
		end
	elseif config.main=="Texts" then Me:TextButton(button,parse,parm)					-- Buttons im Men TEXTS
	else
		master:Debug("invalid button click "..Me:tell(button))
	end
end

-- --------------------------------------------------------------------------
-- EDITBOX CHANGED (something typed)
-- --------------------------------------------------------------------------
function Me:EditboxChange(editbox,parm2)
	if guidebug then master:Debug("GUI EditboxChange "..Me:tell(editbox)..","..Me:tell(parm2)) end
	local parse={strsplit("_",editbox:GetName())} -- z.B. PrattleFrame,Texts,7,Text
	local nr=tonumber(parse[#parse-1])
	local text=editbox:GetText()
	if config.main=="Texts" then Me:TextEditbox(nr,text,editbox)
	elseif config.main=="Replacements" then Me:ReplacementsEditbox(nr,text,editbox)
	end
end

-- --------------------------------------------------------------------------
-- SLIDE
-- --------------------------------------------------------------------------
function Me:Slide(slider)
	local name={strsplit("_",slider:GetName())}
	local item=name[#name-1]
	local newvalue=format("%u",slider:GetValue())
	if guidebug then master:Debug("GUI Slider "..Me:tell(slider)..". value="..slider:GetValue().." formatted="..newvalue) end
	if dbpC.reactOn[item]~=newvalue then
		dbpC.reactOn[item]=tonumber(newvalue)
		getglobal(slider:GetParent():GetName().."_Value"):SetText(newvalue.."%")
		Me:SetUnsavedChanges(true)
	end
end

-- --------------------------------------------------------------------------
-- ROUTE REQUEST
-- --------------------------------------------------------------------------
function Me:Route(request,first,second,third)
	if guidebug then master:Debug("GUI Request "..Me:tell(request)..","..Me:tell(first)..","..Me:tell(second)..","..Me:tell(third)) end
	if unsavedchanges then Me:Unsaved(request,first,second,third)
	elseif request=="CLOSE" then
		if (visiblepanel) then visiblepanel:Hide() end
		visiblepanel=nil
		PrattleFrame:Hide() 	
	else
		if visiblepanel then visiblepanel:Hide() end
		visiblepanel=Me:Widget(request)
		if request~=config.main then
			config={}
			db={}
			dbpC={}
		end
		if request=="Help" then Me:HelpShow()
		elseif request=="Reactions" then Me:ReactionsShow()
		elseif request=="Texts" then Me:TextShow(first,second,third)
		elseif request=="Replacements" then Me:ReplacementsShow()
		end
		local shorthelp=""
		if not L.shorthelp[request] then shorthelp="No help for L.shorthelp["..request.."]"
		else
			for l=1,#L.shorthelp[request],1 do shorthelp=shorthelp.."\n"..L.shorthelp[request][l] end
		end
		Me:Widget("ShortHelp_Text"):SetText(shorthelp)
		visiblepanel:Show()
		PrattleFrame:Show()
		Me:SetUnsavedChanges(false)
	end
end

-- --------------------------------------------------------------------------
-- SHOW HELP
-- --------------------------------------------------------------------------
function Me:HelpShow()
	master:Debug("Keine Hilfe verfgbar")
end

-- --------------------------------------------------------------------------
-- SHOW REACTIONS
-- --------------------------------------------------------------------------
function Me:ReactionsShow()
	if guidebug then master:Debug("GUI ReactionsShow") end
	Me:Clone(PrattleDBpC,dbpC,"reactOn")
	for key,value in pairs(dbpC.reactOn) do
		local title="Reactions_"..key.."_Title"
		local value="Reactions_"..key.."_Value"
		local slide="Reactions_"..key.."_Slide"
		if not Me:Widget(title) then master:Print(gsub(L.msg.noslidefor,"<key>",key),nil,"ERROR")
		else
			Me:Widget(title):SetText(L.title[key] or key)
			Me:Widget(value):SetText(dbpC.reactOn[key].."%")
			Me:Widget(value):SetTextColor(1.0,1.0,1.0,1.0)
			Me:Widget(slide):SetValue(dbpC.reactOn[key])
			Me:Widget(slide):SetValueStep(1)
			Me:Widget(slide.."Low"):SetText(L.text.never)
			Me:Widget(slide.."High"):SetText(L.text.always)
		end
	end
end

-- --------------------------------------------------------------------------
-- SHOW TEXTS
-- --------------------------------------------------------------------------
function Me:TextShow(setfirst,setsecond,setthird)
	if guidebug then master:Debug("GUI TextShow "..Me:tell(setfirst)..","..Me:tell(setsecond)..","..Me:tell(setthird)) end
	-- vorbereiten der Config-Daten fr Darstellung der emotor-Texte
	if not db.emotor then
		Me:Clone(PrattleDB,db,"emotor")
		config={main="Texts",root="emotor",top1=1,top2=1,top3=1,topt=1}
	end
	
	local flip=false
	if not config.first or (setfirst and setfirst~=config.first) then
		local sorted={}
		for list,content in pairs(db[config.root]) do sorted[content.internal.nr]=list end
		config.sorted1=sorted
		config.first=(setfirst or sorted[1])
		config.add1=true
		config.max1=#sorted
		flip=true
	end
	if flip or (setsecond and setsecond~=config.second) then
		config.sorted2={"NOTARGET","UNKNOWN","FRIEND","ENEMY"}
		config.second=(setsecond or "NOTARGET")
		config.add2=true
		config.max2=4
		flip=true
		if not db[config.root][config.first]["NOTARGET"] then  
			config.second="NOCAT"
			config.sorted2={}
			config.add2=false
		end
	end
	if flip or (setthird and setthird~=config.third) then
		config.third=setthird or "say"
		config.sorted3={"emo","say"}
		config.add3=false
		config.top3=1
		config.max3=2
		config.sortedt=db[config.root][config.first][config.second][config.third]
		config.maxt=#config.sortedt
		config.topt=1
	end
	
	-- FIRST-Auswahl Buttons vorbereiten
	local remember_unchanged=unsavedchanges
	for btn=1,99,1 do
		local button=Me:Widget("Texts_Select_First_"..btn)
		local label=config.sorted1[config.top1+btn-1]
		if not button then break
		elseif label then Me:SetWidget(button,db[config.root][label].internal.desc or label,(label==config.first),true)
		elseif config.add1 and config.sorted1[config.top1+btn-2] then Me:SetWidget(button,L.button.NewFirst,false,true)
		else Me:SetWidget(button,"-",false,false)
		end
		button:RegisterForClicks("LeftButtonUp", "RightButtonDown");
		if label==config.first then button:LockHighlight() else button:UnlockHighlight() end
	end
	
	-- SECOND-Auswahl Buttons vorbereiten
	for btn=1,99,1 do
		local button=Me:Widget("Texts_Select_Second_"..btn)
		local label=config.sorted2[config.top2+btn-1]
		if not button then break
		elseif label then Me:SetWidget(button,L.second[label] or label,(label==config.second),true)
		elseif config.add2 and config.sorted2[config.top2+btn-2] then Me:SetWidget(button,L.button.NewSecond,false,true)
		else Me:SetWidget(button,"-",false,false)
		end
		button:RegisterForClicks("LeftButtonUp", "RightButtonUp");
		if label==config.second then button:LockHighlight() else button:UnlockHighlight() end
	end
	-- THIRD-Auswahl Buttons vorbereiten
	for btn=1,99,1 do
		local button=Me:Widget("Texts_Select_Third_"..btn)
		local label=config.sorted3[config.top3+btn-1]
		if not button then break
		elseif label then Me:SetWidget(button,L.third[label] or label,(label==config.third),true)
		elseif config.add3 and config.sorted3[config.top3+btn-2] then Me:SetWidget(button,L.button.NewThird,false,true)
		else Me:SetWidget(button,"-",false,false)
		end
		if label==config.third then button:LockHighlight() else button:UnlockHighlight() end
	end
	-- Texte fllen
	local list=config.sortedt
	config.maxt=#list
	for t=1,999,1 do
		local button=Me:Widget("Texts_"..t.."_Button")
		local editbox=Me:Widget("Texts_"..t.."_Text")
		if not button then break end
		if list[config.topt+t-1] then
			Me:SetWidget(editbox,list[config.topt+t-1].text,editback,true)
			local language=list[config.topt+t-1][UnitName("player")]
			if language then language=GetLanguageByIndex(language) else language=L.text.notselected end
			Me:SetWidget(button,language,(language~=L.text.notselected),true)
		else
			Me:SetWidget(editbox,"",editback,true)
			Me:SetWidget(button,GetLanguageByIndex(defaultlanguage),true,true)
		end
	end
end

function Me:TextButton(button,parse)
	local area=parse[#parse-1]
	local number=parse[#parse]
	if number=="Button" and parse[#parse-2]=="Texts" then
		area="Text"
		number=parse[#parse-1]
	end
	if guidebug then master:Debug("GUI TextButton "..Me:tell(button).." area="..area..", number="..number) end
	if area=="First" 		then
		local fnr=tonumber(number+config.top1-1)
		local req1=config.sorted1[fnr]
		local req2="NOCAT"
		if req1 then
			if db[config.root][req1][config.second] then req2=config.second else req2=nil end
		else
			req1="neu" 
			db[config.root][req1]={internal={nr=fnr,desc="Neu"},[req2]={say={},emo={}}}
		end
		Me:Route("Texts",req1,req2,config.third)
	elseif area=="Second" 	then 
		Me:Route("Texts",config.first,config.sorted2[tonumber(number)],config.third)
	elseif area=="Third" 	then 
		Me:Route("Texts",config.first,config.second,config.sorted3[tonumber(number)])
	else -- area="Text"
		local tbl=db[config.root][config.first][config.second][config.third][config.topt+tonumber(number)-1]
		local set=tbl[playername]
		if not set then set=1 else set=set+1 end
		if set>GetNumLanguages() then set=nil end
		if set~=tbl[playername] then Me:SetUnsavedChanges(true) end
		tbl[playername]=set
		Me:TextShow()
	end
end

function Me:TextEditbox(nr,text,editbox)
	local dbnr=config.topt+nr-1
	local before=db[config.root][config.first][config.second][config.third][dbnr]
	if guidebug then master:Debug("GUI TextEditbox "..Me:tell(editbox).." "..nr.."=db"..dbnr.."\nb="..Me:tell(before).."\nt="..Me:tell(text)) end
	if before then before=before.text
	else before="" end
	if not db[config.root][config.first][config.second][config.third][dbnr] then 
		if text~="" then db[config.root][config.first][config.second][config.third][dbnr]={text=text} end
	else
		if text~="" then db[config.root][config.first][config.second][config.third][dbnr].text=text
		else db[config.root][config.first][config.second][config.third][dbnr]=nil end
	end
	if (text or "")~=before then Me:SetUnsavedChanges(true) end
end

-- --------------------------------------------------------------------------
-- SHOW REPLACEMENTS
-- --------------------------------------------------------------------------
function Me:ReplacementsShow()
	if guidebug then master:Debug("GUI ReplacementShow") end
	-- vorbereiten der Config-Daten fr Darstellung der HeSheIt-Tabelle
	if not db.HeSheIt then
		Me:Clone(PrattleDB,db,"HeSheIt")
		config={main="Replacements",root="HeSheIt",topt=1}
	end
	Me:Widget("Replacements_Head_Key"):SetText(L.text.keyword)
	for t=1,3,1 do Me:Widget("Replacements_Head_"..t):SetText(masterL.sex[t]) end
	
	local sorted={}
	local nr=0
	for key,repl in pairs(db.HeSheIt) do
		nr=nr+1
		sorted[nr]=key
	end
	
	for run=1,#sorted,1 do
		local change=false
		for n=1,#sorted-1,1 do
			if strlower(sorted[n])>strlower(sorted[n+1]) then
				local sortedn=sorted[n]
				sorted[n]=sorted[n+1]
				sorted[n+1]=sortedn
				change=true
			end
		end
		if not change then break end
	end
	
	config.sorted=sorted
	config.maxt=#sorted
	for t=1,999,1 do
		local keybox=Me:Widget("Replacements_"..t.."_Key")
		local key=config.sorted[config.topt+t-1]
		if not keybox then break end
		if key then
			Me:SetWidget(keybox,key,editback,true)
			for s=1,3,1 do Me:SetWidget(Me:Widget("Replacements_"..t.."_"..s),db.HeSheIt[key][s],editback,true) end
		else
			Me:SetWidget(keybox,"",editback,true)
			for s=1,3,1 do Me:SetWidget(Me:Widget("Replacements_"..t.."_"..s),"",editback,true) end
		end
	end
end

function Me:ReplacementsEditbox(nr,text,editbox)
	local parse={strsplit("_",editbox:GetName())}
	local column=parse[#parse]
	local dbnr=config.topt+nr-1
	local key=config.sorted[dbnr]
	local changed=false
	if not key then
		key=""
		config.sorted[dbnr]=key
		config.maxt=1+dbnr
		db[config.root][key]={"","",""}
	end
	if column=="Key" then
		changed=(text~=key)
		if changed then
			config.sorted[dbnr]=text
			db[config.root][text]={}
			for s=1,3,1 do db[config.root][text][s]=db[config.root][key][s] end
			db[config.root][key]=nil
		end
	else
		changed=(db[config.root][key][tonumber(column)]==text)
		db[config.root][key][tonumber(column)]=text
	end
	if changed then Me:SetUnsavedChanges(true) end
end

-- --------------------------------------------------------------------------
-- UNSAVED 
-- --------------------------------------------------------------------------
function Me:SetUnsavedChanges(trueorfalse)
	if guidebug then master:Debug("GUI SetUnsavedChanges "..Me:tell(trueorfalse)) end
	unsavedchanges=trueorfalse
	local savebutton=getglobal(mainpanel.."_Button_Save")
	if unsavedchanges then savebutton:Show() else savebutton:Hide() end
end

-- SHOW SAVE/IGNORE/ABORT 
function Me:Unsaved(request)
	if guidebug then master:Debug("GUI Unsaved "..Me:tell(request)) end
	Me:Widget("Disable"):Show()
	queuedRequest=request
	Me:Widget("Unsaved_Text"):SetText(L.text.unsaved)
	Me:Widget("Unsaved_Save"):SetText(L.button.Save)
	Me:Widget("Unsaved_Ignore"):SetText(L.button.Ignore)
	Me:Widget("Unsaved_Abort"):SetText(L.button.Abort)
	Me:Widget("Unsaved"):Show()
end
-- FINISH UNSAVED
function Me:FinishUnsaved(mode)
	if guidebug then master:Debug("GUI FinishUnsaved "..Me:tell(mode)) end
	Me:Widget("Unsaved"):Hide()
	Me:Widget("Disable"):Hide()
	local request=queuedRequest
	queuedRequest=nil
	if mode=="SAVE" then 
		Me:Save(db,PrattleDB) 
		Me:Save(dbpC,PrattleDBpC) 
	end
	if mode~="ABORT" then Me:SetUnsavedChanges(false) end
	if request and mode~="ABORT" then Me:Route(request) end
end

-- --------------------------------------------------------------------------
-- SAVE
-- --------------------------------------------------------------------------
function Me:Save(tempDB,permDB)
	local nr=0
	for key,value in pairs(tempDB) do
		if type(key)=="number" and type(value)=="table" then
			nr=nr+1
			key=nr
		end
		if not value then -- skip this
		elseif type(value)=="table" then 
			if guidebug then master:Debug("SAVE recurse save on key="..key.."... type(key)="..type(key))	end
			if not permDB[key] then permDB[key]={} end
			Me:Save(value,permDB[key])
		else
			if guidebug then master:Debug("SAVE key="..Me:tell(key)..", value='"..(value or "nil".."'")) end
			permDB[key]=value
		end
	end
end

-- --------------------------------------------------------------------------
-- COMFORT ROUTINES
-- --------------------------------------------------------------------------
function Me:Clone(source,target,tbl)
	if source[tbl] and not target[tbl] then target[tbl]={} end
	for key,data in pairs(source[tbl]) do
		if type(source[tbl][key])=="table" then 
			Me:Clone(source[tbl],target[tbl],key)
		else
			target[tbl][key]=data
		end
	end
end

function Me:Widget(subname) return getglobal(mainpanel.."_"..subname) end

function Me:SetWidget(widget,text,hilight,visible)
	widget:SetText(text)
	if type(hilight)=="boolean" then
	else
		widget.texture=widget:CreateTexture()
		widget.texture:SetAllPoints(widget)
		widget.texture:SetTexture(hilight[1],hilight[2],hilight[3],hilight[4])
	end
	if visible then widget:Show() else widget:Hide() end
end

function Me:tell(variable)
	local result=variable
	if variable==nil then result="nil"
	elseif type(variable)=="boolean" then
		if variable then result="true" else result="false" end
	elseif type(variable)=="table" then 
		if variable[0] then result=variable:GetName() end
		if not variable[0] or not result then
			result=""
			for key,data in pairs(variable) do result=result..","..key end
			result="{"..strsub(result,2).."}"
		end
	end
	if result==nil then result="error-nil" end
	return result
end